/* main.c - commandline and script parsing and such
 * Copyright (C) 2000 Rik Snel <snel@phys.uu.nl>
 * Parts copyright by Eric S. Raymond and the Free Software Foundation
 *
 * This file is part of newsleech.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include<stdio.h>
#include<stdlib.h>
#include<varargs.h>
#include<string.h>
#include<getopt.h>
#include<errno.h>

#include "config.h"
#include "newsleech.h"
#include "getopt.h"
#include "main.h"
#include "visits.h"

const char *exec_name;
int verbose = 0, dummy = 0;
int un_alias = 1;
char defaultaction = 's';
char *defaultcommand = "ARTICLE";
char *defaultfilename = NULL, *defaultheader = NULL;
char *defaultpipeto = NULL;
char *aliasfile = "newsleech.alias";
struct rootnodeinfo rootnode; /* There is ONE rootnode. */

void print_usage(void) {
	printf("\
Simple usage: %s [-s SERVER] -g GROUP [-b|-e|-l HEADER] -r RANGE\n\
", exec_name);
}

void print_help(void) {
	printf("Newsleech %s is a non-interactive rfc977 news retriever.\n", VERSION);
	print_usage();
	printf("\nOptions: (the order of the options does matter)\n");
	printf("\
  -V,  --version		display the version of NewsLeech and exit.\n\
  -h,  --help			print this help.\n\
  -v,  --verbose=LEVEL		be verbose. (bitwise 0-15, default: 0)\n\
  -s,  --server=SERVER		name of the newsserver.\n\
  -p,  --port=PORTNUM		tcp port to connect to. (default: 119)\n\
  -i,  --interval=INTERVAL	retry '400|502 too many connections' (seconds)\n\
  -R,  --aretry=NUMBER          retry same article NUMBER times (default: 0)\n\
       --username=USERNAME	username for servers that require authentication\n\
       --password=PASSWORD	password for servers that require authentication\n\
  -g,  --group=GROUP		name of the newsgroup.\n\
  -r,  --range=RANGE		range of message numbers to operate on.\n\
  -a,  --article		dump the whole message (this is the default).\n\
  -b,  --body			dump only the body of the message.\n\
  -e,  --header			dump only the header of the message.\n\
  -l,  --list=HEADER		get a list of headerlines.\n\
       --stdout                 dump to stdout (this is the default).\n\
  -f,  --filename=FILE		dump to FILE instead of stdout.\n\
  -t,  --pipeto=COMMAND         pipe to stdin of COMMAND instead of stdout.\n\
       --dummy			scan the commandline and exit.\n\
       --script=SCRIPT		execute SCRIPT.\n\
       --aliasfile=PATH		specify alternate alias file.\n\
");
}

/* Fill posting with articles */
int fill_posting(struct postinginfo *posting, char *optarg) {
	char *comp = optarg, *tok, junk, dash;
	int c, first, last;
	verbose_ka("range", optarg);
	if (strcmp("XHDR", posting->command) == 0) {
		if (strcmp(optarg, "all") == 0) {
			posting->articles = realloc(posting->articles, 2*sizeof(int));
			posting->articles[0] = -1;
			posting->count = 2;
			return(0);
		}
		c = sscanf(optarg, "%d%c%d%c", &first, &dash, &last, &junk);
		if (c == 4 || c == 0) return(-1);
		if (c == 3) {
			if (last <= first || dash != '-') return(-1);
			posting->articles = realloc(posting->articles, 2*sizeof(int));
			posting->articles[1] = last;
			posting->articles[0] = first;
			posting->count = 2;
			return(0);
		} 
		if (c == 2) {
			if (dash != '-') return(-1);
			posting->articles = realloc(posting->articles, 2*sizeof(int));
			posting->articles[1] = -1;
			posting->articles[0] = first;
			posting->count = 2;
			return(0);
		}
		posting->articles = realloc(posting->articles, sizeof(int));
		posting->articles[0] = first;
		posting->count = 1;
	} else {
		while ((tok = strtok(comp, ","))) {
			comp = NULL;
			c = sscanf(tok, "%d%c%d%c", &first, &dash, &last, &junk);
			if (c == 4 || c == 0 || c == 2) return(-1);
			if (c == 1) {
				last = first;
				}
			if (last < first) return(-1);
			if (c != 1 && dash != '-') return(-1);
			for (c = first; c <= last; c++) {
				posting->articles = realloc(posting->articles, (posting->count+1)*sizeof(int));
				posting->articles[posting->count] = c;
				posting->count += 1;
			}
		}
	}
	return(0);
}

/* Fill group with postings, because filename is given on commandline */
int fill_group(char *command, struct groupinfo *group, char *filename, char *header, char action, char *pipeto) {
	verbose_k("#here starts a new posting");
	group->postings = realloc(group->postings, (group->count+1)*sizeof(struct postinginfo));
	group->postings[group->count].filename = filename;
	group->postings[group->count].command = command;
	group->postings[group->count].header = header;
	group->postings[group->count].count = 0; /* no articles */
	group->postings[group->count].articles = NULL;
	group->postings[group->count].action = action;
	group->postings[group->count].pipeto = pipeto;
	group->count += 1;
	return(0);
}

int de_alias(char *optarg) {
	FILE *fp;
	if (un_alias) {
		if ((fp = openalias())) {
			push_buffer_state(fp);
			if (!searchalias("aliases", optarg)) {
				pop_buffer_state();
				return(1);
			}
			pop_buffer_state();
		}
	}
	return(0);
}

/* Fill server with groups */
int fill_server(struct serverinfo *server, char *optarg) {
	/* Let us check if *optarg is an alias */
	if (de_alias(optarg)) return(0);
	verbose_ka("group", optarg);
	server->groups = realloc(server->groups, (server->count+1)*sizeof(struct groupinfo));
	server->groups[server->count].name = optarg;
	server->groups[server->count].count = 0; /* no postings */
	server->groups[server->count].postings = NULL;
	server->groups[server->count].first = 0;
	server->groups[server->count].last = 0;
	server->groups[server->count].total = 0;
	server->count += 1;
	return(0);
}

int option_shell(struct rootnodeinfo *rootnode, char *optarg) {
	if (rootnode->count == 0) {
		rootnode->shellout=optarg;
		verbose_ka("shell", optarg);
		return(0);
	}
	fprintf(stderr, "shellout used in wrong place\n");
	exit(1);
	}
	
/* Fill rootnode with servers */
int fill_rootnode(struct rootnodeinfo *rootnode, char *optarg) {
	/* Let us check if *optarg is an alias */
	if (de_alias(optarg)) return(0);
	if (rootnode->count == 0) 
		verbose_k("#defaults are above this line");
	verbose_ka("server", optarg);
	/* Get some memory for our new server. */
	rootnode->servers = realloc(rootnode->servers, (rootnode->count+1)*sizeof(struct serverinfo));
	/* FIXME: should check memfree */
	rootnode->servers[rootnode->count].name = optarg;
	/* Perhaps check if name makes sense? */
	rootnode->servers[rootnode->count].port = DEFAULTPORT;
	rootnode->servers[rootnode->count].interval = DEFAULTINTERVAL;
	rootnode->servers[rootnode->count].aretry = DEFAULTARETRY;
	rootnode->servers[rootnode->count].count = 0; /* no groups */
	rootnode->servers[rootnode->count].groups = NULL;
	rootnode->servers[rootnode->count].user = NULL;
	rootnode->servers[rootnode->count].pass = NULL;
	rootnode->count += 1;
	return(0);
} 
	 
int ch_port(int port, struct rootnodeinfo *rootnode) {
	if (rootnode->servers[rootnode->count-1].count == 0) {
		rootnode->servers[rootnode->count-1].port = port;
		verbose_kd("port", port);
		return(0);
	}
	fprintf(stderr, "Please specify the alternate portnumber before you specify groups.\n");
	exit(1);
}

int ch_interval(int interval, struct rootnodeinfo *rootnode) {
	if (rootnode->servers[rootnode->count-1].count == 0) {
		rootnode->servers[rootnode->count-1].interval = interval;
		verbose_kd("interval", interval);
		return(0);
	}
	fprintf(stderr, "Please specify the retry-interval before you specify groups.\n");
	exit(1);
}

int ch_aretry(int aretry, struct rootnodeinfo *rootnode) {
	if (rootnode->servers[rootnode->count-1].count == 0) {
		rootnode->servers[rootnode->count-1].aretry = aretry;
		verbose_kd("aretry", aretry);
		return(0);
	}
	fprintf(stderr, "Please specify the article-retry before you specify groups.\n");
	exit(1);
}

int ch_user(char *user, struct rootnodeinfo *rootnode) {
	if (rootnode->servers[rootnode->count-1].count == 0) {
		rootnode->servers[rootnode->count-1].user = user;
		verbose_ka("username", user);
		return(0);
	}
	fprintf(stderr, "Please specify the server before you specify the username.\n");
	exit(1);
}

int ch_pass(char *pass, struct rootnodeinfo *rootnode) {
	if (rootnode->servers[rootnode->count-1].count == 0) {
		rootnode->servers[rootnode->count-1].pass =pass;
		verbose_k("password xxxxxxxx");
		return(0);
	}
	fprintf(stderr, "Please specify the username before you specify the password.\n");
	exit(1);
}

int main(int argc, char **argv) {
	int option, option_index, error, c, d, e;
	static struct option long_options[] = {
		{"version", 0, 0, 'V'},
		{"help", 0, 0, 'h'},
		{"verbose", 1, 0, 'v'},
		{"server", 1, 0, 's'},
		{"port", 1, 0, 'p'},
		{"interval", 1, 0, 'i'},
		{"aretry", 1, 0, 'R'},
		{"group", 1, 0, 'g'},
		{"range", 1, 0, 'r'},
		{"body", 0, 0, 'b'},
		{"header", 0, 0, 'e'},
		{"article", 0, 0, 'a'},
		{"list", 1, 0, 'l'},
		{"stdout", 0, 0, '2'},
		{"filename", 0, 0, 'f'},
		{"pipeto", 1, 0, 't'},
		{"dummy", 0, 0, '5'},
		{"script", 1, 0, 1},
		{"username", 1, 0, '3'},
		{"shellout", 1, 0, '6'},
		{"password", 1, 0, '4'},
		{"aliasfile", 1, 0, '7'},
		{0, 0, 0, 0}
	};
	rootnode.count = 0; /* There are 0 servers, currently */
	rootnode.servers = NULL;
	rootnode.shellout = NULL;
	/* Construct the executable name without the leading slash '/' */
	/* directly taken from wget source */
	exec_name = strrchr(argv[0], '/');
	if (!exec_name) exec_name = argv[0];
	else ++exec_name;
	/* Read the options and scripts */
	while (1) {
		option = getopt_long(argc, argv, "-Vhv:s:p:g:r:beal:f:i:t:", 
			long_options, &option_index);
		if (option == -1) break;
		incredibly_large_and_ugly_switch_statement(option, optarg);
		}
	if (optind < argc) {
		fprintf(stderr, "Unrecognized options: ");
		while (optind < argc) printf("%s ", argv[optind++]);
		printf("\n");
		print_usage();
		exit(2);
	}
	if (rootnode.count == 0) {
		yyin = stdin;
		parsescript("standard input", 1);
	}
	if (dummy == 1) verbose_k("dummy");
	verbose_k("end");
	error = 0;
	if (verbose & 4) fprintf(stderr, "Sanitychecking the commandline/script...\n");
	if (rootnode.count == 0) {
		fprintf(stderr, "-no server or group.\n");
		error += 1;
	} else {
		for (c = 0; c < rootnode.count; c++) {
			if (rootnode.servers[c].count == 0) {
				fprintf(stderr, "-no group for %s.\n", rootnode.servers[c].name);
				error += 1;
			} else {
				for (d = 0; d < rootnode.servers[c].count; d++) {
					if (rootnode.servers[c].groups[d].count == 0) {
						fprintf(stderr, "-no filename/range for group %s on server %s\n", rootnode.servers[c].groups[d].name, rootnode.servers[c].name);
						error += 1;
					} else {
						for (e = 0; e < rootnode.servers[c].groups[d].count; e++) {
							if (rootnode.servers[c].groups[d].postings[e].count == 0) {
								fprintf(stderr, "-no range for filename %s in group %s on server %s\n", rootnode.servers[c].groups[d].postings[e].filename, rootnode.servers[c].groups[d].name, rootnode.servers[c].name);
								error += 1;
							}
						}
					}
				}
			}
		}
	}
	if (error != 0) {
		fprintf(stderr, "There were %d error(s) sanitychecking the commandline/script.\n", error);
		exit(1);
	}
	if (verbose & 4) fprintf(stderr, "The commandline/script seems sane.\n");
if (dummy) exit(0);
	if (rootnode.shellout != NULL) {
		system(rootnode.shellout);
	}
	error = visit_servers(rootnode.servers, rootnode.count);
	
	exit(-error);
}	

/* This is an attempt to guess if some string contains whitespace */
int whitespace(char *string) {
	if ( strchr(string, ' ') != NULL ) return(1);
	if ( strchr(string, '\t') != NULL ) return(1);
	return(0);
}

/* Verbose; keyword decimal */
void verbose_kd(char *keyword, int decimal) {
	if ( !(verbose & 4) ) return;
	fprintf(stderr, "> %s %d\n", keyword, decimal);
}

/* Verbose; keyword/comment and argument
 * This function autmomatically prints quotes if argument
 * contains whitespace
 */
void verbose_ka(char *keyword, char *argument) {
	if ( !(verbose & 4) ) return;
	if (whitespace(argument) == 1) {
		fprintf(stderr, "> %s \"%s\"\n", keyword, argument);
		return;
	}
	fprintf(stderr, "> %s %s\n", keyword, argument);
}

/* Verbose; keyword/comment */
void verbose_k(char *keyword) {
	if ( !(verbose & 4) ) return;
	fprintf(stderr, "> %s\n", keyword);
}

/* open alias-file and return FILE* */
FILE *openalias()
{
	FILE *fp;
	char path_to_config[128];
	fp = fopen(aliasfile, "r");
	if (fp) return(fp);
	if (getenv("HOME") == NULL) return(NULL);
	strncpy(path_to_config, getenv("HOME"), 100);
	fp = fopen(strcat(path_to_config, "/.newsleech.alias"), "r");
	return(fp);
}
	
/* stream must be in global yyin */
int searchalias(char *name, char *alias) {
	int option, scripton = 1, linenum = 1;
	char *optarg;
	/*verbose_ka("#we are currently scanning", name);
	verbose_ka("#for alias", alias);*/
	while (scripton == 1) {
		switch(yylex(&linenum, &option, &optarg)) {
			case 0:
				scripton = 2;
				break;
			case 2:
				fprintf(stderr, "Keyword requires argument in line %d of %s.\n", linenum, name);
				exit(1);
				break;
			case 3:
				fprintf(stderr, "Parse error in line %d of %s.\n", linenum, name);
				exit(1);
	 			break;
			case 1:
				/*printf("DEBUG: %c %s\n", option, optarg);*/
				if (option == '9' && strcmp(optarg, alias) == 0) {
					/*verbose_ka("#found alias entry for", alias);*/
					un_alias = 0;
					parsescript(name, linenum);
					un_alias = 1;
					scripton = 0;
				}
			case 4:
		}
		
	}
	if (scripton == 2) {
		/*verbose_ka("#didn't find alias entry for", alias);*/
		return(1);
	}
	return(0);
}

/* stream must be in global yyin */			
void parsescript(char *name, int linenum) {
	int scripton = 1, option;
	char *optarg;
	verbose_ka("#we are currently parsing", name);
	
	while (scripton == 1) {
		switch(yylex(&linenum, &option, &optarg)) {
			case 0:
			case 4:
				scripton=0;
				break;
			case 3:
				fprintf(stderr, "Parse error in line %d of %s.\n", linenum, name);
				exit(1);
	 			break;
			case 2:
				fprintf(stderr, "Keyword requires argument in line %d of %s.\n", linenum, name);
				exit(1);
			case 1:
				if (option == '9' && un_alias == 0) {
					scripton = 0;
				} else {
					incredibly_large_and_ugly_switch_statement(option, optarg);
				}
		}
	}
	verbose_ka("#finished processing", name);
}			

void option_verbose(char *optarg, int optind) {
	if (verbose != 0) {
		fprintf(stderr, "You may only give --verbose LEVEL once.\n");
		exit(1);
	}
	sscanf(optarg, "%d", &verbose);
	if ((optind == 3 || optind == 2) && verbose & 4) { 
		fprintf(stderr, "Parsing commandline, results in scriptform.\n");
	} else {
		if (verbose & 4) fprintf(stderr, "Verbose parsing can only be enabled if -v LEVEL is the first option.\n");
		verbose &= 1+2+8; 
	}
	verbose_k("#verbose is not a real script command");
	verbose_kd("verbose", verbose);
}

void option_port(char *optarg) {
	int port;
	if (sscanf(optarg, "%d", &port) != 1) {
		fprintf(stderr,"Error parsing portnumber.\n");
		exit(1);
	}
	if (SERVERS == 0) {
		fprintf(stderr, "You must specify a server before you can change the portnumber.\n");
		exit(1);
	} else 	ch_port(port, &rootnode);
}

void option_interval(char *optarg) {
	int interval;
	if (sscanf(optarg, "%d", &interval) != 1) {
		fprintf(stderr, "The interval you specified is not understood by this program.\n");
		exit(1);
	}
	if (SERVERS == 0) {
		fprintf(stderr, "You must specify a server before you can change the interval.\n");
		exit(1);
	} else ch_interval(interval, &rootnode);
}

void option_aretry(char *optarg) {
	int aretry;
	if (sscanf(optarg, "%d", &aretry) != 1) {
		fprintf(stderr, "The number of article retries is not understood by this program.\n");
		exit(1);
	}
	if (SERVERS == 0) {
		fprintf(stderr, "You must specify a server before you can change the interval.\n");
		exit(1);
	} else ch_aretry(aretry, &rootnode);
}

void option_username(char *optarg) {
	if (SERVERS == 0) {
		fprintf(stderr, "You must specify a server before you can set the username.\n");
		exit(1);
	}
	ch_user(optarg, &rootnode);
}

void option_password(char *optarg) {
	if (SERVERS == 0) {
		fprintf(stderr, "You must specify a server before you can set the password.\n");
		exit(1);
	}
	ch_pass(optarg, &rootnode);
}

void option_group(char *optarg) {
	if (rootnode.count >= 1) {
		fill_server(&SERVERT, optarg);
	} else {
		if (getenv("NNTPSERVER")) {
			fill_rootnode(&rootnode, getenv("NNTPSERVER"));
			fill_server(&SERVERT, optarg);
		} else {
			fprintf(stderr, "You did not specify a server.\n");
			exit(1);
		}
	}
}

void option_filename(char *optarg) {
	if (SERVERS == 0) {
		defaultaction = 'f';
		defaultfilename = optarg;
	} else {
		if (GROUPS < 1) {
			fprintf(stderr, "You must first specify a group before you cat set the filename.\n");
			exit(1);
		}
		if (POSTINGS == 0) { 
			fill_group(defaultcommand, &GROUPT, optarg, defaultheader, 'f', defaultpipeto);
		} else {
			if (ARTICLES != 0) {
				fill_group(defaultcommand, &GROUPT, optarg, defaultheader, 'f', defaultpipeto);
			} else {	
				POSTINGT.filename = optarg;
				POSTINGT.action = 'f';
			}
		}
	}
	verbose_ka("filename", optarg);
}

void option_pipeto(char *optarg) {
	if (SERVERS == 0) {
		defaultaction = 'p';
		defaultpipeto = optarg;
	} else {
		if (GROUPS < 1) {
			fprintf(stderr, "You must first specify a group before you cat set the filename.\n");
			exit(1);
		}
		if (POSTINGS == 0) { 
			fill_group(defaultcommand, &GROUPT, defaultfilename, defaultheader, 'p', optarg);
		} else {
			if (ARTICLES != 0) {
				fill_group(defaultcommand, &GROUPT, defaultfilename, defaultheader, 'p', optarg);
			} else {	
				POSTINGT.pipeto = optarg;
				POSTINGT.action = 'p';
			}
		}
	}
	verbose_ka("pipeto", optarg);
}

void option_bea(char *scriptcmd, char *tmpcmd){
	if (SERVERS == 0) {
		defaultcommand = tmpcmd;
	} else {
		if (GROUPS < 1) {
			fprintf(stderr, "You must first specify a group.\n");
			exit(1);
		} 
		if (POSTINGS == 0) {
			fill_group(tmpcmd, &GROUPT, defaultfilename, defaultheader, defaultaction, defaultpipeto);	
		} else {
			if (ARTICLES != 0) {
				fill_group(tmpcmd, &GROUPT, defaultfilename, defaultheader, defaultaction, defaultpipeto);
			} else {
				POSTINGT.command = tmpcmd;
			}
		}
	}
	verbose_k(scriptcmd);
}

void option_list(char *optarg) {
	if (SERVERS == 0) {
		defaultcommand = "XHDR";
		defaultheader = optarg;
	} else {
		if (GROUPS < 1) {
			fprintf(stderr, "You must first specify a group.\n");
			exit(1);
		}
		if (POSTINGS == 0) {
			fill_group("XHDR", &GROUPT, defaultfilename, optarg, defaultaction, defaultpipeto);
		} else {
			if (ARTICLES != 0) {
				fill_group("XHDR", &GROUPT, defaultfilename, optarg, defaultaction, defaultpipeto);
			} else {
				POSTINGT.header = optarg;
				POSTINGT.command = "XHDR";
			}		
		}
	}
	verbose_ka("list", optarg);
}

void option_stdout(void) {
	if (SERVERS == 0) {
		defaultaction = 's';
		verbose_k("stdout");
	} else {
		if (GROUPS < 1) {
			fprintf(stderr, "You must first specify a group before you cat set the filename.\n");
			exit(1);
		}
		if (POSTINGS == 0) { 
			fill_group(defaultcommand, &GROUPT, defaultfilename, defaultheader, 's', defaultpipeto);
			verbose_k("stdout");
		} else {
			if (ARTICLES != 0) {
				fill_group(defaultcommand, &GROUPT, defaultfilename, defaultheader, 's', defaultpipeto);
				verbose_k("stdout");
			} else {	
				verbose_k("stdout");
				POSTINGT.action = 's';
			}
		}
	}
}

void option_range(char *optarg) {
	if (SERVERS < 1) {
		fprintf(stderr, "You did not specify a server or a group.\n");
		exit(1);
	}
	if (GROUPS < 1) {
		fprintf(stderr, "You did not specify a group.\n");
		exit(1);
	}
	if (POSTINGS < 1) {
		fill_group(defaultcommand, &GROUPT, defaultfilename, defaultheader, defaultaction, defaultpipeto);
	}
	if (ARTICLES != 0) {
		fill_group(defaultcommand, &GROUPT, defaultfilename, defaultheader, defaultaction, defaultpipeto);
		/*fprintf(stderr, "Are you sure this is what you want; -r 12 -r 13 is NOT the same as -r 12,13.\n");*/
		/*exit(1);*/
	}
	if (fill_posting(&POSTINGT, optarg) < 0) {
		fprintf(stderr, "Error interpreting RANGE.\n");
		exit(1);
	}
}

void option_script(char *optarg) {
	FILE *fp;
	fp = fopen(optarg, "r");
	if (fp==NULL) {
		fprintf(stderr, "Error opening %s for reading.\n", optarg);
		exit(1);
	}
	push_buffer_state(fp);
	//yyin = fp;
	parsescript(optarg, 1);
	pop_buffer_state();
	fclose(fp);
}

int incredibly_large_and_ugly_switch_statement(int option, char *optarg) {
	char *tmpcmd, *scriptcmd;
	switch (option) {
		case 'V':
			printf("\
Newsleech %s, Copyright (C) 2000 Rik Snel <snel@phys.uu.nl>.\n\
Parts copyright by Eric S. Raymond and the Free Software Foundation.\n\n\
This program is free software; you can redistribute it and/or modify it\n\
under the terms of the GNU General Public License as publised by the Free\n\
Software Foundation; either version 2 of the License, or (at your option)\n\
any later version. This program comes with ABSOLUTELY NO WARRANTY.\n\
", VERSION);
			exit(0);
			break;
		case 'h':
			print_help();
			exit(0);
			break;
		case 'v':
			option_verbose(optarg, optind);
			break;
		case 'p':
			option_port(optarg);
			break;
		case 'i':
			option_interval(optarg);
			break;
		case 'R':
			option_aretry(optarg);
			break;
		case '3':
			option_username(optarg);
			break;
		case '4':
			option_password(optarg);
			break;
		case '6':
			option_shell(&rootnode, optarg);
			break;
		case 's':
			fill_rootnode(&rootnode, optarg);
			break;
		case 'g':
			option_group(optarg);
			break;
		case 'f':
			option_filename(optarg);
			break;
		case 'r':
			option_range(optarg);
			break;
		case 'b':
			tmpcmd = "BODY";
			scriptcmd = "body";
		case 'e':
			if (option == 'e') {
				tmpcmd = "HEAD";
				scriptcmd = "head";
			}
		case 'a':
			if (option == 'a') {
				tmpcmd = "ARTICLE";
				scriptcmd = "article";
			}
			option_bea(scriptcmd, tmpcmd);
			break;
		case 'l':
			option_list(optarg);
			break;
		case 't':
			option_pipeto(optarg);
			break;
		case '5':
			dummy = 1;
			break;
		case 1:
			option_script(optarg);
			break;
		case '7':
			aliasfile = optarg;
		case '2':
			option_stdout();
			break;
		case '?':
			print_usage();
			exit(2);
		case '9':
			fprintf(stderr, "Thou shalt not use the 'alias'-statement in an ordinary script!\n");
			exit(1);
		default:
			fprintf(stderr, "Not implemented.\n");
			exit(0);
			break;
	}
	return(0);
}
